import { ResolverFactory } from 'enhanced-resolve';
import { FileProcessor } from './file_processor';
import Resolver = require('enhanced-resolve/lib/Resolver');
export declare type Path = string;
export declare type FileToDeps = Map<Path, Set<Path>>;
export declare type ResolverOptionsGenerator = () => Partial<ResolverFactory.ResolverOption>;
/**
 * If reference is a glob pattern (for example `./foo/*.ts`), this function should transform to
 * the list of files, such as `['./foo/bar.ts', './foo/baz.ts']`.
 */
export declare type ReferenceTransformFn = (reference: string, sourceFile: string) => string[] | string;
export declare class DependencyTree {
    private readonly rootDirs;
    private readonly resolver;
    private readonly ignoreGlobs;
    readonly transformReference: ReferenceTransformFn;
    private readonly fileProcessors;
    /**
     * @param rootDirs The root directories to search for files to build the dependency tree
     * @param resolver A resolver to be used for resolving references between files
     * @param ignoreGlobs An array of glob patterns that will be use to exclude files from the
     *     dependency tree on match
     * @param transformReference A function that can be used to transform a reference into a
     *     different path
     */
    constructor(rootDirs: Path[], resolver?: Resolver, ignoreGlobs?: string[], transformReference?: ReferenceTransformFn);
    /**
     * Returns dependencies (and transitive dependencies) of a given set of files
     *
     * @param fileToDeps A mapping between files and their dependencies
     * @param files The files to get dependencies of
     * @returns Any file (indirectly) depending on the given files
     */
    static getDependencies(fileToDeps: FileToDeps, files: Path[]): Set<Path>;
    /**
     * Returns references (and transitive references) to a given set of files
     *
     * @param fileToDeps A mapping between files and their dependencies
     * @param files The files to get references for
     * @returns Any file (indirectly) referencing the given files
     */
    static getReferences(fileToDeps: FileToDeps, files: Path[]): Set<Path>;
    /**
     * Walk a dependency graph to find all the files related to the entrypoint files. Removes the
     * entrypoints from the set of visited nodes since it does not make sense for a file to be
     * "related" to itself.
     *
     * @param graph the graph to walk
     * @param files the entrypoints to start the walk on
     */
    private static findRelatedFiles;
    /**
     * Creates a synchronous resolver
     *
     * @param generator
     * @returns the resolver
     */
    static createResolver(generator?: ResolverOptionsGenerator): Resolver;
    /**
     * Adds an additional file processor.
     * Processors are called in the order they were added.
     */
    addFileProcessor(fileProcessor: FileProcessor): void;
    /**
     * Walks through all found files and builds a map of their resolved dependencies
     * and a set of missing dependencies.
     * The map returned maps from a file (absolute path) to its referenced dependencies (absolute
     * paths)
     *
     * Built-in (Node) and package dependencies are ignored.
     */
    gather({ filesToProcess, allFilesToProcess, batchSize, }?: {
        batchSize?: number;
        filesToProcess?: readonly string[];
        allFilesToProcess?: readonly string[];
    }): Promise<{
        missing: FileToDeps;
        resolved: FileToDeps;
    }>;
    /**
     * This resolves a referenced module and adds it to either the resolved files or the set of
     * non-existent dependencies
     *
     * @param file The file to resolve from (e.g. the file containing the reference)
     * @param referencedModule The module to resolve (e.g. the import/require target)
     * @param importedModules The set to add a module to that can be resolved
     * @param missingModules The map of sets to add a module to that can not
     *     be resolved
     */
    resolveAndCollect(file: Path, referencedModule: string[] | string, importedModules: Set<Path>, missingModules: FileToDeps): void;
    /**
     * Returns all imported files for a given file
     *
     * @param file The file to get the imported files for
     * @param missing A set that will be populated with modules that couldn't be
     *     resolved
     * @param files All files found in this dependency run.
     * @returns {Set<string>}
     */
    private getImportedFiles;
    /**
     * @returns {string[]} An array of supported file extensions
     */
    getSupportedFileTypes(): string[];
    /**
     * Globs all supported files from the {rootDirs}
     *
     * @returns {string[]} An array of found files (absolute paths)
     */
    getFiles(): readonly Path[];
}
