"use strict";
// Copyright 2021 Canva Inc. All Rights Reserved.
Object.defineProperty(exports, "__esModule", { value: true });
exports.DependencyTree = void 0;
const bim = require("builtin-modules");
const enhanced_resolve_1 = require("enhanced-resolve");
const fg = require("fast-glob");
const fs = require("fs");
const path = require("path");
const logger_1 = require("./logger");
const css_1 = require("./processors/css");
const directive_1 = require("./processors/directive");
const typescript_1 = require("./processors/typescript");
const SyncAsyncFileSystemDecorator = require("enhanced-resolve/lib/SyncAsyncFileSystemDecorator");
const promise_pool_1 = require("@supercharge/promise-pool");
// This is a set of built-in modules, e.g. `path`, `fs`, etc.
const builtinModules = new Set(bim);
const info = logger_1.debug.extend('info');
const error = logger_1.debug.extend('error');
const debug = logger_1.debug.extend('debug');
class DependencyTree {
    /**
     * @param rootDirs The root directories to search for files to build the dependency tree
     * @param resolver A resolver to be used for resolving references between files
     * @param ignoreGlobs An array of glob patterns that will be use to exclude files from the
     *     dependency tree on match
     * @param transformReference A function that can be used to transform a reference into a
     *     different path
     */
    constructor(rootDirs, resolver = DependencyTree.createResolver(), ignoreGlobs = ['**/node_modules/**'], transformReference = (ref) => ref) {
        this.rootDirs = rootDirs;
        this.resolver = resolver;
        this.ignoreGlobs = ignoreGlobs;
        this.transformReference = transformReference;
        this.fileProcessors = [];
        for (const dir of this.rootDirs) {
            this.addFileProcessor(new typescript_1.TypeScriptFileProcessor(dir));
        }
        this.addFileProcessor(new css_1.CssFileProcessor());
        this.addFileProcessor(new directive_1.TSDirectiveProcessor());
    }
    /**
     * Returns dependencies (and transitive dependencies) of a given set of files
     *
     * @param fileToDeps A mapping between files and their dependencies
     * @param files The files to get dependencies of
     * @returns Any file (indirectly) depending on the given files
     */
    static getDependencies(fileToDeps, files) {
        const graph = new DirectedGraph(fileToDeps);
        return DependencyTree.findRelatedFiles(graph, files);
    }
    /**
     * Returns references (and transitive references) to a given set of files
     *
     * @param fileToDeps A mapping between files and their dependencies
     * @param files The files to get references for
     * @returns Any file (indirectly) referencing the given files
     */
    static getReferences(fileToDeps, files) {
        // invert the graph because we are interested in the files that *reference* the entrypoint
        // files, rather than the files that the entrypoint files references (which is what
        // `DependencyTree.getDependencies` does)
        const graph = new DirectedGraph(fileToDeps).transpose();
        return DependencyTree.findRelatedFiles(graph, files);
    }
    /**
     * Walk a dependency graph to find all the files related to the entrypoint files. Removes the
     * entrypoints from the set of visited nodes since it does not make sense for a file to be
     * "related" to itself.
     *
     * @param graph the graph to walk
     * @param files the entrypoints to start the walk on
     */
    static findRelatedFiles(graph, files) {
        const visited = new Set();
        for (const referencingFile of graph.manyWalkDfs(files)) {
            visited.add(referencingFile);
        }
        // delete all the entrypoints since it does not make sense for a file to reference or
        // depend on itself
        for (const file of files) {
            visited.delete(file);
        }
        return visited;
    }
    /**
     * Creates a synchronous resolver
     *
     * @param generator
     * @returns the resolver
     */
    static createResolver(generator = () => ({})) {
        const opts = {
            ...generator(),
            fileSystem: new SyncAsyncFileSystemDecorator(new enhanced_resolve_1.CachedInputFileSystem(new enhanced_resolve_1.NodeJsInputFileSystem(), 4000)),
        };
        return enhanced_resolve_1.ResolverFactory.createResolver(opts);
    }
    /**
     * Adds an additional file processor.
     * Processors are called in the order they were added.
     */
    addFileProcessor(fileProcessor) {
        this.fileProcessors.push(fileProcessor);
    }
    /**
     * Walks through all found files and builds a map of their resolved dependencies
     * and a set of missing dependencies.
     * The map returned maps from a file (absolute path) to its referenced dependencies (absolute
     * paths)
     *
     * Built-in (Node) and package dependencies are ignored.
     */
    async gather({ 
    // The files to gather dependencies for
    // by default globbing is used to find files from rootDirs
    filesToProcess, 
    // All files collected
    // by default globbing is used to find files from rootDirs
    allFilesToProcess, batchSize = 10, } = {}) {
        if (batchSize < 1) {
            throw new Error(`expected a batch size greater than 0, got '${batchSize}'`);
        }
        const fileToDeps = new Map();
        const missing = new Map();
        const allFiles = allFilesToProcess !== null && allFilesToProcess !== void 0 ? allFilesToProcess : this.getFiles();
        const files = filesToProcess !== null && filesToProcess !== void 0 ? filesToProcess : allFiles;
        info(`Found a total of ${files.length} source files`);
        const getDepsForFilesTasks = files.map((file) => async () => {
            info('Scanning %s', file);
            const importedFiles = await this.getImportedFiles(file, missing, allFiles);
            fileToDeps.set(file, importedFiles);
        });
        await runInBatches(getDepsForFilesTasks, batchSize);
        return {
            missing,
            resolved: fileToDeps,
        };
    }
    /**
     * This resolves a referenced module and adds it to either the resolved files or the set of
     * non-existent dependencies
     *
     * @param file The file to resolve from (e.g. the file containing the reference)
     * @param referencedModule The module to resolve (e.g. the import/require target)
     * @param importedModules The set to add a module to that can be resolved
     * @param missingModules The map of sets to add a module to that can not
     *     be resolved
     */
    resolveAndCollect(file, referencedModule, importedModules, missingModules) {
        if (Array.isArray(referencedModule)) {
            for (const m of referencedModule) {
                this.resolveAndCollect(file, m, importedModules, missingModules);
            }
            return;
        }
        if (builtinModules.has(referencedModule)) {
            debug('%s is a built-in module, ignoring', referencedModule);
            return;
        }
        try {
            debug('trying to resolve %s against %s via enhanced-resolve', referencedModule, file);
            const resolved = this.resolver.resolveSync({}, path.dirname(file), referencedModule);
            if (resolved) {
                importedModules.add(resolved);
                return;
            }
        }
        catch (e) {
            // We write the message to stderr but treat the module as missing
            error(e.message);
        }
        if (referencedModule) {
            error("Couldn't find: %o", referencedModule);
            const missingFileDeps = missingModules.get(file) || new Set();
            missingFileDeps.add(referencedModule);
            missingModules.set(file, missingFileDeps);
        }
    }
    /**
     * Returns all imported files for a given file
     *
     * @param file The file to get the imported files for
     * @param missing A set that will be populated with modules that couldn't be
     *     resolved
     * @param files All files found in this dependency run.
     * @returns {Set<string>}
     */
    async getImportedFiles(file, missing, files) {
        const contents = fs.readFileSync(file, 'utf8');
        const allImported = new Set();
        let matched = false;
        for (const fileProcessor of this.fileProcessors) {
            if (fileProcessor.match(file)) {
                matched = true;
                const importedFiles = await fileProcessor.process(file, contents, missing, files, this);
                for (const imported of Array.from(importedFiles)) {
                    allImported.add(imported);
                }
            }
        }
        if (!matched) {
            /* istanbul ignore next */
            throw new Error(`No file processor matches ${file}`);
        }
        return allImported;
    }
    /**
     * @returns {string[]} An array of supported file extensions
     */
    getSupportedFileTypes() {
        const supportedFileTypes = new Set(this.fileProcessors.reduce((acc, processor) => [...acc, ...processor.supportedFileTypes()], []));
        return Array.from(supportedFileTypes);
    }
    /**
     * Globs all supported files from the {rootDirs}
     *
     * @returns {string[]} An array of found files (absolute paths)
     */
    getFiles() {
        const supportedFileTypes = this.getSupportedFileTypes();
        return this.rootDirs.reduce((acc, rootDir) => {
            return acc.concat(fg.sync(`**/*.{${supportedFileTypes.join(',')}}`, {
                cwd: rootDir,
                ignore: this.ignoreGlobs,
                absolute: true,
            }));
        }, []);
    }
}
exports.DependencyTree = DependencyTree;
async function runInBatches(tasks, batchSize) {
    await promise_pool_1.PromisePool.withConcurrency(batchSize)
        .for(tasks)
        .process((task) => task());
}
/**
 * Directed graph. Internally represented using the edges of the graph. For example, the following
 * directed graph:
 *
 * <pre>
 *        +-> b --> c
 *    a --|
 *        +-> d
 * </pre>
 *
 * Can be represented using the following sequence of pairs, representing each of the edges in the
 * graph:
 *
 * <pre>
 *   (a, b)
 *   (a, d)
 *   (b, c)
 * </pre>
 */
class DirectedGraph {
    constructor(edges) {
        this.edges = edges;
    }
    /**
     * Computes the transpose of the dependency graph & return a graph of files to referencing files.
     *
     * Visualise this transformation as a reversal of all (directed) edges in the graph, e.g. from
     * this
     *
     * <pre>
     *       +-> b --> c
     *   a --|
     *       +-> d
     * </pre>
     *
     * to this
     *
     * <pre>
     *   c --> b --+
     *             |-> a
     *         d --+
     * </pre>
     *
     * @param fileToDeps the graph to invert
     */
    transpose() {
        const invertedEdges = new Map();
        for (const [node, endpointNodes] of this.edges) {
            for (const endpointNode of endpointNodes) {
                if (invertedEdges.has(endpointNode)) {
                    assertNonNull(invertedEdges.get(endpointNode), 'expected outNodes to not be null').add(node);
                }
                else {
                    invertedEdges.set(endpointNode, new Set([node]));
                }
            }
        }
        return new DirectedGraph(invertedEdges);
    }
    /**
     * Perform multiple depth-first walks of the graph, starting each walk at one of the provided
     * entrypoints.
     *
     * @param entrypoints start points for each of the depth-first walks
     */
    *manyWalkDfs(entrypoints) {
        const visited = new Set();
        for (const entrypoint of entrypoints) {
            yield* this.walkDfs(entrypoint, visited);
        }
    }
    /**
     * Walk the graph in a depth-first manner.
     *
     * @param entrypoint the starting node for the walk
     * @param visited the nodes that have already been visited. injected to ensure that nodes are only
     *        visited once across multiple walks
     */
    *walkDfs(entrypoint, visited = new Set()) {
        const toVisit = new Stack();
        toVisit.add(entrypoint);
        while (!toVisit.isEmpty()) {
            const node = toVisit.next();
            if (visited.has(node)) {
                continue;
            }
            yield node;
            visited.add(node);
            const outNodes = this.edges.get(node);
            if (outNodes) {
                toVisit.addMany(outNodes);
            }
        }
    }
}
/** O(1) unbounded stack ADT */
class Stack {
    constructor() {
        this.stack = [];
    }
    add(val) {
        this.stack.push(val);
    }
    addMany(vals) {
        for (const val of vals) {
            this.add(val);
        }
    }
    next() {
        return assertNonNull(this.stack.pop(), 'tried to pop from an empty stack');
    }
    isEmpty() {
        return this.stack.length === 0;
    }
}
function assertNonNull(val, message) {
    if (val == null) {
        throw new Error(message);
    }
    return val;
}
